/* SRCDrop - Filename: AUDIO.C
 *
 * Function: Essentially provides all the wave input and output routines.
 *
 *   Copyright (c) 2005 John Edwards <john.edwards33@ntlworld.com>
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Library General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Portions, (c) Michael Smith <msmith@labyrinth.net.au>
 * and libvorbis examples, (c) Monty <monty@xiph.org>
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <io.h>
#include <fcntl.h>
#include <malloc.h>
#include <math.h>
#include "audio.h"
#include "i18n.h"
#include "resample.h"
#include "misc.h"

#include "../src/float_cast.h"

/* Macros to read header data */
#define READ_U32_LE(buf) \
	(((buf)[3]<<24)|((buf)[2]<<16)|((buf)[1]<<8)|((buf)[0]&0xff))

#define READ_U16_LE(buf) \
	(((buf)[1]<<8)|((buf)[0]&0xff))

#define READ_U32_BE(buf) \
	(((buf)[0]<<24)|((buf)[1]<<16)|((buf)[2]<<8)|((buf)[3]&0xff))

#define READ_U16_BE(buf) \
	(((buf)[0]<<8)|((buf)[1]&0xff))

static unsigned char pcm_guid[16] =
{
	/* (00000001-0000-0010-8000-00aa00389b71) */

	0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x10, 0x00, 0x80, 0x00, 0x00, 0xaa, 0x00, 0x38, 0x9b, 0x71
};

static unsigned char ieee_float_guid[16] =
{
	/* (00000003-0000-0010-8000-00aa00389b71) */

	0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x10, 0x00, 0x80, 0x00, 0x00, 0xaa, 0x00, 0x38, 0x9b, 0x71
};

/* Define the supported formats here */
input_format formats[] = {
	{wav_id, 12, wav_open, wav_close, "wav", N_("WAV file reader")},
	{NULL, 0, NULL, NULL, NULL, NULL}
};

input_format *open_audio_file(FILE *in, src_proc_opt *opt)
{
	int j=0;
	unsigned char *buf=NULL;
	int buf_size=0, buf_filled=0;
	int size, ret;

	while (formats[j].id_func) {
		size = formats[j].id_data_len;
		if (size >= buf_size) {
			buf = realloc(buf, size);
			buf_size = size;
		}

		if (size > buf_filled) {
			ret = fread(buf+buf_filled, 1, buf_size-buf_filled, in);
			buf_filled += ret;

			if (buf_filled < size) {
				/* File truncated */
				j++;
				continue;
			}
		}

		if (formats[j].id_func(buf, buf_filled)) {
			/* ok, we now have something that can handle the file */
			if (formats[j].open_func(in, opt, buf, buf_filled)) {
				free(buf);
				return &formats[j];
			}
		}
		j++;
	}

	free(buf);

	return NULL;
}

static int seek_forward(FILE *in, int length)
{
	if (fseek(in, length, SEEK_CUR)) {
		/* Failed. Do it the hard way. */
		unsigned char buf[1024];
		int seek_needed = length, seeked;
		while (seek_needed > 0) {
			seeked = fread(buf, 1, seek_needed>1024?1024:seek_needed, in);
			if (!seeked)
				return 0; /* Couldn't read more, can't read file */
			else
				seek_needed -= seeked;
		}
	}
	return 1;
}


static int find_wav_chunk(FILE *in, char *type, unsigned int *len, int to_log_file)
{
	unsigned char buf[8];

	while (1) {
		if (fread(buf,1,8,in) < 8) {
			/* Suck down a chunk specifier */
			error_handler("Warning: Unexpected EOF in reading WAV header\n");
			return 0; /* EOF before reaching the appropriate chunk */
		}

		if (memcmp(buf, type, 4)) {
			*len = READ_U32_LE(buf+4);
			if (!seek_forward(in, *len))
				return 0;

			buf[4] = 0;
			if (to_log_file)
				error_handler("Skipping chunk of type \"%s\", length %d\n", buf, *len);
		}
		else {
			*len = READ_U32_LE(buf+4);
			return 1;
		}
	}
}

double read_IEEE80(unsigned char *buf)
{
	int s=buf[0]&0xff;
	int e=((buf[0]&0x7f)<<8)|(buf[1]&0xff);
	double f=((unsigned long)(buf[2]&0xff)<<24)| ((buf[3]&0xff)<<16)| ((buf[4]&0xff)<<8) | (buf[5]&0xff);

	if (e==32767) {
		if (buf[2]&0x80)
			return HUGE_VAL; /* Really NaN, but this won't happen in reality */
		else {
			if (s)
				return -HUGE_VAL;
			else
				return HUGE_VAL;
		}
	}

	f=ldexp(f,32);
	f+= ((buf[6]&0xff)<<24)| ((buf[7]&0xff)<<16)| ((buf[8]&0xff)<<8) | (buf[9]&0xff);

	return ldexp(f, e-16446);
}

int wav_id(unsigned char *buf, int len)
{
	unsigned int flen;
	
	if (len<12) return 0; /* Something screwed up */

	if (memcmp(buf, "RIFF", 4))
		return 0; /* Not wave */

	flen = READ_U32_LE(buf+4); /* We don't use this */

	if (memcmp(buf+8, "WAVE",4))
		return 0; /* RIFF, but not wave */

	return 1;
}

int wav_open(FILE *in, src_proc_opt *opt, unsigned char *oldbuf, int buflen)
{
	unsigned char buf[41];
	unsigned int len;
	int samplesize;
	wav_fmt format;
	wavfile *wav = malloc(sizeof(wavfile));

	/* Ok. At this point, we know we have a WAV file. Now we have to detect
	 * whether we support the subtype, and we have to find the actual data
	 * We don't (for the wav reader) need to use the buffer we used to id this
	 * as a wav file (oldbuf)
	 */
	if (!find_wav_chunk(in, "fmt ", &len, opt->logerr))
		return 0; /* EOF */

	if (len < 16) {
		error_handler("Warning: Unrecognised format chunk in WAV header\n");
		return 0; /* Weird format chunk */
	}

	/* A common error is to have a format chunk that is not 16 or 18 bytes
	 * in size.  This is incorrect, but not fatal, so we only warn about 
	 * it instead of refusing to work with the file.  Please, if you
	 * have a program that's creating format chunks of sizes other than
	 * 16, 18 or 40 bytes in size, report a bug to the author.
	 * (40 bytes accommodates WAVEFORMATEXTENSIBLE conforming files.)
	 */
	if (len!=16 && len!=18 && len!=40)
		error_handler("Warning: INVALID format chunk in wav header.\n"
				" Trying to read anyway (may not work)...\n");

	/* Deal with stupid broken apps. Don't use these programs.
	 */
	if (fread(buf,1,len,in) < len) {
		error_handler("Warning: Unexpected EOF in reading WAV header\n");
		return 0;
	}

	format.format =      READ_U16_LE(buf); 
	format.channels =    READ_U16_LE(buf+2); 
	format.samplerate =  READ_U32_LE(buf+4);
	format.bytespersec = READ_U32_LE(buf+8);
	format.align =       READ_U16_LE(buf+12);
	format.samplesize =  READ_U16_LE(buf+14);

	if (format.format == WAVE_FORMAT_PCM) {
		samplesize = format.samplesize / 8;
		opt->format = samplesize;
		opt->read_samples = wav_read;
		wav->channel_map = UNDEFINED;
		opt->channel_mask = 0;
	}
	else if (format.format == WAVE_FORMAT_IEEE_FLOAT) {
		samplesize = 4;
		opt->format = 5;
		opt->read_samples = wav_ieee_read;
		wav->channel_map = UNDEFINED;
		opt->channel_mask = 0;
	}
	else if (format.format == WAVE_FORMAT_EXTENSIBLE) {
		format.channel_mask = READ_U32_LE(buf+20);
		wav->channel_map = DEFINED;
		opt->channel_mask = format.channel_mask;
		if (memcmp(buf+24, pcm_guid, 16) == 0) {
			samplesize = format.samplesize / 8;
			opt->format = samplesize;
			opt->read_samples = wav_read;
		}
		else if (memcmp(buf+24, ieee_float_guid, 16) == 0) {
			samplesize = 4;
			opt->format = 5;
			opt->read_samples = wav_ieee_read;
		}
		else {
			error_handler("ERROR: Wav file is unsupported type (must be standard PCM\n"
					" or type 3 floating point PCM)(2)\n");
			return 0;
		}
	}
	else {
		error_handler("ERROR: Wav file is unsupported type (must be standard PCM\n"
				" or type 3 floating point PCM)(1)\n");
		return 0;
	}

	if (!find_wav_chunk(in, "data", &len, opt->logerr))
		return 0; /* EOF */

	if(format.align != format.channels * samplesize) {
		/* This is incorrect according to the spec. Warn loudly, then ignore
		 * this value.
		 */
		error_handler("Warning: WAV 'block alignment' value is incorrect, ignoring.\n" 
		    "The software that created this file is incorrect.\n");
	}
	if(format.samplesize == samplesize * 8 && 
		(format.samplesize == 32 || format.samplesize == 24 ||
		 format.samplesize == 16 || format.samplesize == 8)) {
		/* OK, good - we have a supported format,
		 * now we want to find the size of the file
		 */
		opt->rate = format.samplerate;
		opt->channels = format.channels;

		wav->f = in;
		wav->samplesread = 0;
		wav->bigendian = 0;
		wav->channels = format.channels; /* This is in several places. The price
							of trying to abstract stuff. */
		wav->samplesize = format.samplesize;
		opt->samplesize = format.samplesize;

		if (len)
			opt->total_samples_per_channel = len/(format.channels*samplesize);
		else {
			unsigned int pos;
			pos = ftell(in);
			if (fseek(in, 0, SEEK_END) == -1)
				opt->total_samples_per_channel = 0; /* Give up */
			else {
				opt->total_samples_per_channel = (ftell(in) - pos)/(format.channels*samplesize);
				fseek(in,pos, SEEK_SET);
			}
		}
		wav->totalsamples = opt->total_samples_per_channel;

		opt->readdata = (void *)wav;

		return 1;
	}
	else {
		error_handler("ERROR: Wav file is unsupported subformat (must be 8, 16, 24 or 32 bit PCM\n"
				"or floating point PCM)\n");
		return 0;
	}
}


long wav_read(void *in, float *buffer, int samples)
{

	wavfile *f = (wavfile *)in;
	int sampbyte = f->samplesize / 8;
	signed char *buf = alloca(samples*sampbyte*f->channels);
	long bytes_read = fread(buf, 1, samples*sampbyte*f->channels, f->f);
	int i;
	long realsamples;

	if (f->totalsamples && f->samplesread + bytes_read/(sampbyte*f->channels) > f->totalsamples) {
		bytes_read = sampbyte*f->channels*(f->totalsamples - f->samplesread);
	}

	realsamples = bytes_read/(sampbyte*f->channels);
	f->samplesread += realsamples;
		
	if (f->samplesize==8) {
		unsigned char *bufu = (unsigned char *)buf;
		for (i = 0; i < realsamples; i++)
			buffer[i] = ((int)(bufu[i]) - 128) / 128.0f;
	}
	else if (f->samplesize==16) {
		short *short_sample_buffer = (short *)buf;
		for (i = 0; i < realsamples*f->channels; i++)
			buffer[i] = ((buf[i * 2 + 1]<<8) |
					(buf[i * 2] & 0xff)) / 32768.0f;
	}
	else if (f->samplesize==24) {
		int *int_sample_buffer = (int *)buf;
		for (i = 0; i < realsamples*f->channels; i++)
			buffer[i] = 	 	  ((buf[i * 3 + 2] << 16) |
				(((unsigned char *)buf)[i * 3 + 1] << 8) |
				(((unsigned char *)buf)[i * 3] & 0xff)) 
				/ 8388608.0f;
	}
	else if (f->samplesize==32) {
		int *int_sample_buffer = (int *)buf;
		for (i = 0; i < realsamples*f->channels; i++)
			buffer[i] = 		  ((buf[i * 4 + 3] << 24) |
				(((unsigned char *)buf)[i * 4 + 2] << 16) |
				(((unsigned char *)buf)[i * 4 + 1] << 8) |
				(((unsigned char *)buf)[i * 4] & 0xff)) 
				/ 2147483648.0f;
	}
	else {
		error_handler("Internal error: attempt to read unsupported "
			"bitdepth %d\n", f->samplesize);
		return 0;
	}

	return realsamples;
}

long wav_ieee_read(void *in, float *buffer, int samples)
{
	wavfile *f = (wavfile *)in;
	float *buf = alloca(samples*4*f->channels); /* de-interleave buffer */
	long bytes_read = fread(buf,1,samples*4*f->channels, f->f);
	int i;
	long realsamples;


	if (f->totalsamples && f->samplesread + bytes_read/(4*f->channels) > f->totalsamples)
		bytes_read = 4*f->channels*(f->totalsamples - f->samplesread);
	realsamples = bytes_read/(4*f->channels);
	f->samplesread += realsamples;

	for (i=0; i < realsamples * f->channels; i++)
		buffer[i] = buf[i];

	return realsamples;
}

void wav_close(void *info)
{
	wavfile *f = (wavfile *)info;

	free(f);
}

/*
 * W A V E   O U T P U T
 */

output_audio_file *open_output_audio_file(src_proc_opt *opt)
{
	output_audio_file *aufile = malloc(sizeof(output_audio_file));

	aufile->outputFormat = opt->format;

	aufile->samplerate = opt->resamplefreq;
	aufile->channels = opt->channels;
	aufile->samples = 0;

	switch (opt->format) {
		case WAV_FMT_8BIT:
			aufile->bits_per_sample = 8;
			break;
		case WAV_FMT_16BIT:
			aufile->bits_per_sample = 16;
			break;
		case WAV_FMT_24BIT:
			aufile->bits_per_sample = 24;
			break;
		case WAV_FMT_32BIT:
		case WAV_FMT_FLOAT:
			aufile->bits_per_sample = 32;
			break;
		default:
			if (aufile)
				free(aufile);
			return NULL;
	}

	aufile->sndfile = fopen(opt->filename, "wb");

	if (aufile->sndfile == NULL) {
		if (aufile)
			free(aufile);
		return NULL;
	}

	if (opt->channel_mask)
		write_wav_extensible_header(aufile, opt);
	else
		write_wav_header(aufile, opt);

	return aufile;
}

int write_audio_file(output_audio_file *aufile, void *sample_buffer, int samples)
{
	char *buf = (char *)sample_buffer;
	switch (aufile->outputFormat) {
		case WAV_FMT_8BIT:
			return write_audio_8bit(aufile, buf, samples);
		case WAV_FMT_16BIT:
			return write_audio_16bit(aufile, buf, samples);
		case WAV_FMT_24BIT:
			return write_audio_24bit(aufile, buf, samples);
		case WAV_FMT_32BIT:
			return write_audio_32bit(aufile, buf, samples);
		case WAV_FMT_FLOAT:
			return write_audio_float(aufile, buf, samples);
		default:
			return 0;
	}

	return 0;
}

void close_audio_file(output_audio_file *aufile, src_proc_opt *opt)
{
	fseek(aufile->sndfile, 0, SEEK_SET);
	if (opt->channel_mask)
		write_wav_extensible_header(aufile, opt);
	else
		write_wav_header(aufile, opt);

	if(opt)
		free(opt);

	fclose(aufile->sndfile);

	if (aufile)
		free(aufile);
}

#define WRITE_U32(buf, x) *(buf)     = (unsigned char)((x)&0xff);\
                          *((buf)+1) = (unsigned char)(((x)>>8)&0xff);\
                          *((buf)+2) = (unsigned char)(((x)>>16)&0xff);\
                          *((buf)+3) = (unsigned char)(((x)>>24)&0xff);

#define WRITE_U16(buf, x) *(buf)     = (unsigned char)((x)&0xff);\
                          *((buf)+1) = (unsigned char)(((x)>>8)&0xff);

static int write_wav_header(output_audio_file *aufile, src_proc_opt *opt)
{
	unsigned char char_buf[5];

	unsigned short channels   = aufile->channels;
	unsigned int samplerate   = aufile->samplerate;
	unsigned int bytespersec  = aufile->channels * aufile->samplerate * aufile->bits_per_sample / 8;
	unsigned short align      = aufile->channels * aufile->bits_per_sample / 8;
	unsigned short samplesize = aufile->bits_per_sample;
	unsigned int size         = aufile->samples * (aufile->bits_per_sample / 8) < 0x7fffffff ?
					aufile->samples * (aufile->bits_per_sample / 8) : 0x7fffffff;	

	strcpy(char_buf, "RIFF");
	fwrite(char_buf, 4, 1, aufile->sndfile);
	WRITE_U32(char_buf,  size + 44 - 8);
	fwrite(char_buf, 4, 1, aufile->sndfile);
	strcpy(char_buf, "WAVE");
	fwrite(char_buf, 4, 1, aufile->sndfile);
	strcpy(char_buf, "fmt ");
	fwrite(char_buf, 4, 1, aufile->sndfile);
	WRITE_U32(char_buf,  16);
	fwrite(char_buf, 4, 1, aufile->sndfile);
	if (aufile->outputFormat == WAV_FMT_FLOAT) {
		WRITE_U16(char_buf,  3);
	}
	else {
		WRITE_U16(char_buf,  1);
	}
	fwrite(char_buf, 2, 1, aufile->sndfile);
	WRITE_U16(char_buf,  channels);
	fwrite(char_buf, 2, 1, aufile->sndfile);
	WRITE_U32(char_buf,  samplerate);
	fwrite(char_buf, 4, 1, aufile->sndfile);
	WRITE_U32(char_buf,  bytespersec);
	fwrite(char_buf, 4, 1, aufile->sndfile);
	WRITE_U16(char_buf,  align);
	fwrite(char_buf, 2, 1, aufile->sndfile);
	WRITE_U16(char_buf,  samplesize);
	fwrite(char_buf, 2, 1, aufile->sndfile);
	strcpy(char_buf, "data");
	fwrite(char_buf, 4, 1, aufile->sndfile);
	WRITE_U32(char_buf,  size);
	fwrite(char_buf, 4, 1, aufile->sndfile);

	return 1;
}

static int write_wav_extensible_header(output_audio_file *aufile, src_proc_opt *opt)
{
	unsigned char header[68];
	unsigned char* p = header;
	unsigned int bytes = (aufile->bits_per_sample + 7) / 8;
	float data_size = (float)bytes * aufile->samples;
	unsigned long channelMask;
	unsigned long word32;

	channelMask = opt->channel_mask;

	*p++ = 'R'; *p++ = 'I'; *p++ = 'F'; *p++ = 'F';

	word32 = (data_size + (68 - 8) < (float)MAXWAVESIZE) ?
        	(unsigned long)data_size + (68 - 8)  :  (unsigned long)MAXWAVESIZE;
	*p++ = (unsigned char)(word32 >>  0);
	*p++ = (unsigned char)(word32 >>  8);
	*p++ = (unsigned char)(word32 >> 16);
	*p++ = (unsigned char)(word32 >> 24);

	*p++ = 'W'; *p++ = 'A'; *p++ = 'V'; *p++ = 'E';

	*p++ = 'f'; *p++ = 'm'; *p++ = 't'; *p++ = ' ';

	*p++ = /*0x10*/0x28; *p++ = 0x00; *p++ = 0x00; *p++ = 0x00;

	/* WAVE_FORMAT_EXTENSIBLE */
	*p++ = 0xFE; *p++ = 0xFF;

	*p++ = (unsigned char)(aufile->channels >> 0);
	*p++ = (unsigned char)(aufile->channels >> 8);

	word32 = (unsigned long)(aufile->samplerate + 0.5);
	*p++ = (unsigned char)(word32 >>  0);
	*p++ = (unsigned char)(word32 >>  8);
	*p++ = (unsigned char)(word32 >> 16);
	*p++ = (unsigned char)(word32 >> 24);

	word32 = aufile->samplerate * bytes * aufile->channels;
	*p++ = (unsigned char)(word32 >>  0);
	*p++ = (unsigned char)(word32 >>  8);
	*p++ = (unsigned char)(word32 >> 16);
	*p++ = (unsigned char)(word32 >> 24);

	word32 = bytes * aufile->channels;
	*p++ = (unsigned char)(word32 >>  0);
	*p++ = (unsigned char)(word32 >>  8);

	*p++ = (unsigned char)(aufile->bits_per_sample >> 0);
	*p++ = (unsigned char)(aufile->bits_per_sample >> 8);

	/* cbSize */
	*p++ = (unsigned char)(22);
	*p++ = (unsigned char)(0);

	/* WAVEFORMATEXTENSIBLE */

	/* wValidBitsPerSample */
	*p++ = (unsigned char)(aufile->bits_per_sample >> 0);
	*p++ = (unsigned char)(aufile->bits_per_sample >> 8);

	/* dwChannelMask */
	word32 = channelMask;
	*p++ = (unsigned char)(word32 >>  0);
	*p++ = (unsigned char)(word32 >>  8);
	*p++ = (unsigned char)(word32 >> 16);
	*p++ = (unsigned char)(word32 >> 24);

	/* SubFormat */
	if (aufile->outputFormat == WAV_FMT_FLOAT)
	{
		/* KSDATAFORMAT_SUBTYPE_IEEE_FLOAT: 00000003-0000-0010-8000-00aa00389b71 */
		*p++ = 0x03;
		*p++ = 0x00;
		*p++ = 0x00;
		*p++ = 0x00;
		*p++ = 0x00; *p++ = 0x00; *p++ = 0x10; *p++ = 0x00; *p++ = 0x80; *p++ = 0x00;
		*p++ = 0x00; *p++ = 0xaa; *p++ = 0x00; *p++ = 0x38; *p++ = 0x9b; *p++ = 0x71;
	} else {
		/* KSDATAFORMAT_SUBTYPE_PCM: 00000001-0000-0010-8000-00aa00389b71 */
		*p++ = 0x01;
		*p++ = 0x00;
		*p++ = 0x00;
		*p++ = 0x00;
		*p++ = 0x00; *p++ = 0x00; *p++ = 0x10; *p++ = 0x00; *p++ = 0x80; *p++ = 0x00;
		*p++ = 0x00; *p++ = 0xaa; *p++ = 0x00; *p++ = 0x38; *p++ = 0x9b; *p++ = 0x71;
	}

	/* end WAVEFORMATEXTENSIBLE */

	*p++ = 'd'; *p++ = 'a'; *p++ = 't'; *p++ = 'a';

	word32 = data_size < MAXWAVESIZE ?
        	(unsigned long)data_size : (unsigned long)MAXWAVESIZE;
	*p++ = (unsigned char)(word32 >>  0);
	*p++ = (unsigned char)(word32 >>  8);
	*p++ = (unsigned char)(word32 >> 16);
	*p++ = (unsigned char)(word32 >> 24);

	return fwrite(header, sizeof(header), 1, aufile->sndfile);
}

static int write_audio_8bit(output_audio_file *aufile, void *sample_buffer, unsigned int samples)
{
	int              ret;
	unsigned int     i;
	unsigned char    *sample_buffer8 = (unsigned char*)sample_buffer;
	unsigned char    *data = malloc(samples*aufile->bits_per_sample*sizeof(char)/8);

	aufile->samples += samples;

	for (i = 0; i < samples; i++)
		data[i] = (sample_buffer8[i]+128) & 0xFF;

	ret = fwrite(data, samples, aufile->bits_per_sample/8, aufile->sndfile);

	if (data)
		free(data);

	return ret;
}

static int write_audio_16bit(output_audio_file *aufile, void *sample_buffer, unsigned int samples)
{
	int          ret;
	unsigned int i;
	short        *sample_buffer16 = (short*)sample_buffer;
	char         *data = malloc(samples*aufile->bits_per_sample*sizeof(char)/8);

	aufile->samples += samples;

	for (i = 0; i < samples; i++) {
		data[i*2] = (char)(sample_buffer16[i] & 0xFF);
		data[i*2+1] = (char)((sample_buffer16[i] >> 8) & 0xFF);
	}

	ret = fwrite(data, samples, aufile->bits_per_sample/8, aufile->sndfile);

	if (data)
		free(data);

	return ret;
}

static int write_audio_24bit(output_audio_file *aufile, void *sample_buffer, unsigned int samples)
{
	int          ret;
	unsigned int i;
	long         *sample_buffer24 = (long*)sample_buffer;
	char         *data = malloc(samples*aufile->bits_per_sample*sizeof(char)/8);

	aufile->samples += samples;

	for (i = 0; i < samples; i++) {
		data[i*3] = (char)(sample_buffer24[i] & 0xFF);
		data[i*3+1] = (char)((sample_buffer24[i] >> 8) & 0xFF);
		data[i*3+2] = (char)((sample_buffer24[i] >> 16) & 0xFF);
	}

	ret = fwrite(data, samples, aufile->bits_per_sample/8, aufile->sndfile);

	if (data)
		free(data);

	return ret;
}

static int write_audio_32bit(output_audio_file *aufile, void *sample_buffer, unsigned int samples)
{
	int          ret;
	unsigned int i;
	long         *sample_buffer32 = (long*)sample_buffer;
	char         *data = malloc(samples*aufile->bits_per_sample*sizeof(char)/8);

	aufile->samples += samples;

	for (i = 0; i < samples; i++) {
		data[i*4] = (char)(sample_buffer32[i] & 0xFF);
		data[i*4+1] = (char)((sample_buffer32[i] >> 8) & 0xFF);
		data[i*4+2] = (char)((sample_buffer32[i] >> 16) & 0xFF);
		data[i*4+3] = (char)((sample_buffer32[i] >> 24) & 0xFF);
	}

	ret = fwrite(data, samples, aufile->bits_per_sample/8, aufile->sndfile);

	if (data)
		free(data);

	return ret;
}

static int write_audio_float(output_audio_file *aufile, void *sample_buffer, unsigned int samples)
{
	int           ret;
	unsigned int  i;
	float         *sample_buffer_f = (float*)sample_buffer;
	unsigned char *data = malloc(samples*aufile->bits_per_sample*sizeof(char)/8);

	aufile->samples += samples;

	for (i = 0; i < samples; i++) {
		int   exponent,
		      mantissa,
		      negative = 0;
		float in = sample_buffer_f[i];

		data[i*4] = 0;
		data[i*4+1] = 0;
		data[i*4+2] = 0;
		data[i*4+3] = 0;
		if (in == 0.0)
			continue;

		if (in < 0.0) {
			in *= -1.0;
			negative = 1;
		}
		in = (float)frexp(in, &exponent);
		exponent += 126;
		in *= (float)0x1000000;
		mantissa = (((int)in) & 0x7FFFFF);

		if (negative)
			data[i*4+3] |= 0x80;

		if (exponent & 0x01)
			data[i*4+2] |= 0x80;

		data[i*4] = mantissa & 0xFF;
		data[i*4+1] = (mantissa >> 8) & 0xFF;
		data[i*4+2] |= (mantissa >> 16) & 0x7F;
		data[i*4+3] |= (exponent >> 1) & 0x7F;
	}

	ret = fwrite(data, samples, aufile->bits_per_sample/8, aufile->sndfile);

	if (data)
		free(data);

	return ret;
}

void* output_to_PCM(float *input, void *sample_buffer, int channels,
                    int samples, int format)
{
	unsigned int    i;

	char   *char_sample_buffer = (char*)sample_buffer;
	short *short_sample_buffer = (short*)sample_buffer;
	long   *long_sample_buffer = (long*)sample_buffer;
	float *float_sample_buffer = (float*)sample_buffer;

	/*
	 * Copy output to a standard PCM buffer
	 */
	switch (format) {
		case WAV_FMT_8BIT: {
			for (i = 0; i < samples * channels; i++) {
				double value = input[i] * (1.0 * 0x7f);
				if (value > (1.0 * 0x7f)) value = (1.0 * 0x7f);
				if (value < (-8.0 * 0x10)) value = (-8.0 * 0x10);
				char_sample_buffer[i] = lrint(value);
			}
			break;
		}
		case WAV_FMT_16BIT: {
			for (i = 0; i < samples * channels; i++) {
				double value = input[i] * (1.0 * 0x7fff);
				if (value > (1.0 * 0x7fff)) value = (1.0 * 0x7fff);
				if (value < (-8.0 * 0x1000)) value = (-8.0 * 0x1000);
				short_sample_buffer[i] = lrint(value);
			}
			break;
		}
		case WAV_FMT_24BIT:{
			for (i = 0; i < samples * channels; i++) {
				double value = input[i] * (1.0 * 0x7fffff);
				if (value > (1.0 * 0x7fffff)) value = (1.0 * 0x7fffff);
				if (value < (-8.0 * 0x100000)) value = (-8.0 * 0x100000);
				long_sample_buffer[i] = lrint(value);
			}
			break;
		}
		case WAV_FMT_32BIT: {
			for (i = 0; i < samples * channels; i++) {
				double value = input[i] * (1.0 * 0x7fffffff);
				if (value > (1.0 * 0x7fffffff)) value = (1.0 * 0x7fffffff);
				if (value < (-8.0 * 0x10000000)) value = (-8.0 * 0x10000000);
				long_sample_buffer[i] = lrint(value);
			}
			break;
		}
		case WAV_FMT_FLOAT: {
			for (i = 0; i < samples * channels; i++) {
				if (input[i] > 1.0) input[i] = 1.0;
				else if (input[i] < -1.0) input[i] = -1.0;
				float_sample_buffer[i] = input[i];
			}
			break;
		}
	}

	return sample_buffer;
}

/*
 * end of audio.c
 */
